'use strict';

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

var _chai = require('chai');

var _chai2 = _interopRequireDefault(_chai);

var _sinonChai = require('sinon-chai');

var _sinonChai2 = _interopRequireDefault(_sinonChai);

var _index = require('../index');

var _index2 = _interopRequireDefault(_index);

var _deepFreeze = require('deep-freeze');

var _deepFreeze2 = _interopRequireDefault(_deepFreeze);

var _ramdaSrcCompose = require('ramda/src/compose');

var _ramdaSrcCompose2 = _interopRequireDefault(_ramdaSrcCompose);

_chai2['default'].use(_sinonChai2['default']);
var expect = _chai2['default'].expect;

describe('operations', function () {
    var ops = undefined;

    beforeEach(function () {
        ops = (0, _index2['default'])();
    });

    it('wrapBatched', function () {
        var pushFour = ops.push(4);
        var pushFive = ops.push(5);
        var arr = (0, _deepFreeze2['default'])([1, 2, 3]);

        var pusher = ops.batch((0, _ramdaSrcCompose2['default'])(pushFive, pushFour));
        expect(pusher).to.be.a('function');

        var result = pusher(arr);
        expect(result).to.deep.equal([1, 2, 3, 4, 5]);
    });

    describe('object', function () {
        describe('batched mutations', function () {
            it('deepMerges', function () {
                var baseObj = (0, _deepFreeze2['default'])({
                    change: 'Tommi',
                    dontChange: 25,
                    deeper: {
                        dontChange: 'John',
                        change: 30
                    }
                });
                var mergeObj = (0, _deepFreeze2['default'])({
                    change: 'None',
                    add: 'US',
                    deeper: {
                        add: 'US',
                        change: 35
                    }
                });
                var result = undefined;
                var merger = ops.deepMerge(mergeObj);
                ops.batched(function () {
                    result = merger(baseObj);
                    expect((0, _index.canMutate)(result)).to.be['true'];
                    expect((0, _index.canMutate)(result.deeper)).to.be['true'];
                });

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect((0, _index.canMutate)(result.deeper)).to.be['false'];

                expect(result).to.not.equal(baseObj);

                expect(result).to.contain.all.keys(['change', 'dontChange', 'add', 'deeper']);
                expect(result.change).to.not.equal(baseObj.change);
                expect(result.dontChange).to.equal(baseObj.dontChange);

                expect(result.deeper).to.not.equal(baseObj.deeper);
                expect(result.deeper).to.contain.all.keys(['dontChange', 'change', 'add']);
                expect(result.deeper.dontChange).to.equal(baseObj.deeper.dontChange);
                expect(result.deeper.change).to.not.equal(baseObj.deeper.change);
            });

            it('omits a single key', function () {
                var obj = (0, _deepFreeze2['default'])({
                    name: 'Tommi',
                    age: 25
                });

                var result = undefined;
                var omitter = ops.omit('age');

                ops.batched(function () {
                    result = omitter(obj);
                    expect((0, _index.canMutate)(result)).to.be['true'];
                });

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect(result).to.not.contain.keys(['age']);
            });

            it('omits an array of keys', function () {
                var obj = (0, _deepFreeze2['default'])({
                    name: 'Tommi',
                    age: 25
                });

                var result = undefined;

                var omitter = ops.omit(['age']);
                ops.batched(function () {
                    result = omitter(obj);

                    expect((0, _index.canMutate)(result)).to.be['true'];
                });

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect(result).to.not.contain.keys(['age']);
            });

            it('sets a value in path', function () {
                var obj = (0, _deepFreeze2['default'])({
                    first: {
                        second: {
                            value: 'value',
                            maintain: true
                        },
                        maintain: true
                    },
                    maintain: true
                });
                var result = undefined;

                var setter = ops.setIn('first.second.value', 'anotherValue');

                ops.batched(function () {
                    result = setter(obj);

                    expect((0, _index.canMutate)(result)).to.be['true'];
                });

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect(result).not.to.equal(obj);
                expect(result.first.second.value).to.equal('anotherValue');
                expect(result.maintain).to.be['true'];
                expect(result.first.maintain).to.be['true'];
                expect(result.first.second.maintain).to.be['true'];
            });
        });

        describe('immutable ops', function () {
            it('deepMerges', function () {
                var baseObj = (0, _deepFreeze2['default'])({
                    change: 'Tommi',
                    dontChange: 25,
                    deeper: {
                        dontChange: 'John',
                        change: 30
                    }
                });
                var mergeObj = (0, _deepFreeze2['default'])({
                    change: 'None',
                    add: 'US',
                    deeper: {
                        add: 'US',
                        change: 35
                    }
                });

                var merger = ops.deepMerge(mergeObj);
                var result = merger(baseObj);

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect((0, _index.canMutate)(result.deeper)).to.be['false'];

                expect(result).to.not.equal(baseObj);

                expect(result).to.contain.all.keys(['change', 'dontChange', 'add', 'deeper']);
                expect(result.change).to.not.equal(baseObj.change);
                expect(result.dontChange).to.equal(baseObj.dontChange);

                expect(result.deeper).to.not.equal(baseObj.deeper);
                expect(result.deeper).to.contain.all.keys(['dontChange', 'change', 'add']);
                expect(result.deeper.dontChange).to.equal(baseObj.deeper.dontChange);
                expect(result.deeper.change).to.not.equal(baseObj.deeper.change);
            });

            it('omits a single key', function () {
                var obj = (0, _deepFreeze2['default'])({
                    name: 'Tommi',
                    age: 25
                });

                var omitter = ops.omit('age');
                var result = omitter(obj);

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect(result).to.not.contain.keys(['age']);
            });

            it('omits an array of keys', function () {
                var obj = (0, _deepFreeze2['default'])({
                    name: 'Tommi',
                    age: 25
                });

                var omitter = ops.omit(['age']);
                var result = omitter(obj);

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect(result).to.not.contain.keys(['age']);
            });

            it('sets a value in path', function () {
                var obj = (0, _deepFreeze2['default'])({
                    first: {
                        second: {
                            value: 'value',
                            maintain: true
                        },
                        maintain: true
                    },
                    maintain: true
                });

                var setter = ops.setIn('first.second.value', 'anotherValue');

                var result = setter(obj);

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect(result).not.to.equal(obj);
                expect(result.first.second.value).to.equal('anotherValue');
                expect(result.maintain).to.be['true'];
                expect(result.first.maintain).to.be['true'];
                expect(result.first.second.maintain).to.be['true'];
            });
        });
    });

    describe('array', function () {
        describe('batched mutations', function () {
            it('push', function () {
                var push = ops.push;
                var arr = (0, _deepFreeze2['default'])([5, 4]);
                var pusher = push((0, _deepFreeze2['default'])([1, 2, 3]));
                var result = ops.batched(function () {
                    return pusher(arr);
                });

                expect(result).to.not.equal(arr);

                expect(result).to.deep.equal([5, 4, 1, 2, 3]);
            });

            it('insert', function () {
                var insert = ops.insert;
                var arr = (0, _deepFreeze2['default'])([1, 2, 5]);
                var inserter = insert(2, (0, _deepFreeze2['default'])([3, 4]));
                var result = ops.batched(function () {
                    return inserter(arr);
                });

                expect(result).to.deep.equal([1, 2, 3, 4, 5]);
            });

            it('filter', function () {
                var arr = (0, _deepFreeze2['default'])([0, 1, 2, 3]);
                var result = undefined;

                ops.batched(function () {
                    result = ops.filter(function (item) {
                        return item % 2 === 0;
                    }, arr);
                    expect((0, _index.canMutate)(result)).to.be['true'];
                });

                expect(result).to.deep.equal([0, 2]);
                expect((0, _index.canMutate)(result)).to.be['false'];
            });

            it('set', function () {
                var arr = (0, _deepFreeze2['default'])([1, 2, 987, 4]);

                var result = ops.batched(function () {
                    var setter = ops.set(2, 3);
                    var res = setter(arr);
                    expect((0, _index.canMutate)(res)).to.be['true'];
                    return res;
                });

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect(result).to.deep.equal([1, 2, 3, 4]);
            });

            it('splice with deletions', function () {
                var splice = ops.splice;
                var arr = (0, _deepFreeze2['default'])([1, 2, 3, 3, 3, 4]);
                var splicer = splice(2, 2, []);

                var result = ops.batched(function () {
                    return splicer(arr);
                });

                expect(result).to.deep.equal([1, 2, 3, 4]);
            });

            it('splice with additions', function () {
                var splice = ops.splice;
                var arr = (0, _deepFreeze2['default'])([1, 5]);
                var splicer = splice(1, 0, [2, 3, 4]);

                var result = ops.batched(function () {
                    return splicer(arr);
                });

                expect(result).to.deep.equal([1, 2, 3, 4, 5]);
            });
        });

        describe('immutable ops', function () {
            it('push', function () {
                var push = ops.push;
                var arr = (0, _deepFreeze2['default'])([5, 4]);
                var pusher = push((0, _deepFreeze2['default'])([1, 2, 3]));
                var result = pusher(arr);

                expect(result).to.not.equal(arr);

                expect(result).to.deep.equal([5, 4, 1, 2, 3]);
            });

            it('insert', function () {
                var insert = ops.insert;
                var arr = (0, _deepFreeze2['default'])([1, 2, 5]);
                var inserter = insert(2, (0, _deepFreeze2['default'])([3, 4]));
                var result = inserter(arr);

                expect(result).to.deep.equal([1, 2, 3, 4, 5]);
            });

            it('filter', function () {
                var arr = (0, _deepFreeze2['default'])([0, 1, 2, 3]);

                var result = ops.filter(function (item) {
                    return item % 2 === 0;
                }, arr);

                expect(result).to.deep.equal([0, 2]);
                expect((0, _index.canMutate)(result)).to.be['false'];
            });

            it('set', function () {
                var arr = (0, _deepFreeze2['default'])([1, 2, 987, 4]);

                var setter = ops.set(2, 3);
                var result = setter(arr);

                expect((0, _index.canMutate)(result)).to.be['false'];
                expect(result).to.deep.equal([1, 2, 3, 4]);
            });

            it('splice with deletions', function () {
                var splice = ops.splice;
                var arr = (0, _deepFreeze2['default'])([1, 2, 3, 3, 3, 4]);
                var splicer = splice(2, 2, []);

                var result = splicer(arr);
                expect(result).to.deep.equal([1, 2, 3, 4]);
            });

            it('splice with additions', function () {
                var splice = ops.splice;
                var arr = (0, _deepFreeze2['default'])([1, 5]);
                var splicer = splice(1, 0, [2, 3, 4]);

                var result = splicer(arr);

                expect(result).to.deep.equal([1, 2, 3, 4, 5]);
            });
        });
    });
});